{-# LANGUAGE TupleSections #-}

module Cipher ( 
    vigenere,
    parallelVigenere,
    parallelChunkingVigenere
) where

import Data.List(transpose, nub, sort, maximumBy)
import Data.Ord (comparing)
import Data.Char (ord)
import Data.Map (Map, fromListWith, toList, findWithDefault)
import GHC.Conc (par, pseq)
import Control.Parallel.Strategies (
    Strategy,
    evalList,
    using,
    parMap,
    rdeepseq,
    rparWith,
    parListChunk,
    rseq)
import Data.Ratio (denominator)

vigenere :: [Char] -> IO ()
vigenere crypt = do
    let filteredCrypt = filter (/=' ') crypt
        dists = fmap (wrap filteredCrypt) [1..length filteredCrypt `div` 20]
        optimalDist = maximumBy (comparing rate) dists
        key = fmap (deriveCharacter engFreqs) optimalDist
        chars a b = ['A'..'Z'] !! ((ord b - ord a) `mod` 26)
    putStrLn key

parList :: Strategy a -> Strategy [a]
parList strat = evalList (rparWith strat)

parallelVigenere :: [Char] -> IO ()
parallelVigenere crypt = do
    let filteredCrypt = filter (/=' ') crypt
        dists = parMap rdeepseq (wrap filteredCrypt) [1..length filteredCrypt `div` 20]
        optimalDist = maximumBy (comparing parallelRate) dists
            `using` parList rseq
        key = parMap rdeepseq (deriveCharacter engFreqs) optimalDist
    putStrLn key

parallelChunkingVigenere :: [Char] -> IO ()
parallelChunkingVigenere crypt = do
    let filteredCrypt = filter (/=' ') crypt
        dists = parMap rdeepseq (wrap filteredCrypt) [1..length filteredCrypt `div` 20]
        optimalDist = maximumBy (comparing parallelRate) dists
            `using` parListChunk 100 rseq 
        key = parMap rdeepseq (deriveCharacter engFreqs) optimalDist
    putStrLn key

engFreqs :: [Double]
engFreqs = [
    0.081, 0.014, 0.027, 0.042, 0.127, 0.022, 0.020, 0.060, 0.069, 0.001, 
    0.007, 0.040, 0.024, 0.067, 0.075, 0.019, 0.000, 0.059, 0.063, 0.090, 
    0.027, 0.009, 0.023, 0.001, 0.019, 0.000 ] 

aggregate :: Ord a => [a] -> Map a Int
aggregate = fromListWith (+) . fmap (, 1)

avg :: Fractional a => [a] -> a
avg as = sum as / fromIntegral (length as)

parse :: Int -> [a] -> [[a]]
parse _ [] = []
parse n as = 
    let (h, r) = splitAt n as
    in h:parse n r

wrap :: [a] -> Int -> [[a]]
wrap as n = transpose $ parse n as

count :: (Ord a, Fractional b) => [a] -> b
count str = 
    let charCounts = snd <$> toList (aggregate str)
        l = length str
        denominator = fromIntegral $ l * (l - 1)
        numerator = fromIntegral $ sum $ fmap (\c -> c * (c-1)) charCounts
    in numerator / denominator

parallelCount :: (Ord a, Fractional b) => [a] -> b
parallelCount str = 
    let charCounts = snd <$> toList (aggregate str)
        l = length str
        denominator = fromIntegral $ l * (l - 1)
        numerator = fromIntegral $ sum $ parMap rdeepseq (\c -> c * (c-1)) charCounts
    in numerator / denominator

rate :: (Ord a, Fractional b) => [[a]] -> b
rate d =  avg (fmap count d) - fromIntegral (length d) / 3000.0 

parallelRate :: (Ord a, Fractional b) => [[a]] -> b
parallelRate d =  avg (fmap parallelCount d) - fromIntegral (length d) / 3000.0 

sumMult :: Num a => [a] -> [a] -> a
sumMult v0 v1 = sum $ zipWith (*) v0 v1

swap :: Num a => [a] -> [a] -> Char -> a
swap v0 v1 letter = sumMult v0 (drop (ord letter - ord 'A') (cycle v1))  

deriveCharacter :: RealFrac a => [a] -> String -> Char
deriveCharacter possible sample =
    let charCounts = aggregate sample
        samp c = findWithDefault 0 c charCounts
        real = fmap (fromIntegral . samp) ['A'..'Z']
    in maximumBy (comparing $ swap possible real) ['A'..'Z']
